package de.mattv.quarry

import com.mojang.authlib.GameProfile
import mcjty.theoneprobe.api.IProbeInfo
import net.minecraft.core.BlockPos
import net.minecraft.core.BlockPos.MutableBlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.network.chat.TextComponent
import net.minecraft.network.protocol.Packet
import net.minecraft.network.protocol.game.ClientGamePacketListener
import net.minecraft.network.protocol.game.ClientboundBlockEntityDataPacket
import net.minecraft.server.level.ServerLevel
import net.minecraft.util.Tuple
import net.minecraft.world.InteractionHand
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraft.world.item.enchantment.Enchantments
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.entity.BlockEntityTicker
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.MinecraftForge
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.FakePlayer
import net.minecraftforge.common.util.FakePlayerFactory
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.energy.IEnergyStorage
import net.minecraftforge.event.world.BlockEvent
import net.minecraftforge.items.CapabilityItemHandler
import net.minecraftforge.items.ItemHandlerHelper
import java.util.*
import kotlin.math.max
import kotlin.math.min
import kotlin.math.pow
import kotlin.streams.asSequence

object FakeQuarryPlayer {
    private val profile = GameProfile(UUID.fromString("21a819e9-a98d-4e58-bd52-f60b9570a918"), "[MQuarry]")
    fun get(level: ServerLevel): FakePlayer = FakePlayerFactory.get(level, profile)
}

class QuarryTileTicker<T: BlockEntity?> : BlockEntityTicker<T> {
    override fun tick(level: Level, pos: BlockPos, state: BlockState, be: T) {
        if (level.isClientSide()) return
        val tile = be as QuarryTile
        tile.tick(level)
    }
}

class QuarryTile(private val myPos: BlockPos, state: BlockState) : BlockEntity(QuarryMod.QUARRY_TILE.get(), myPos, state), IEnergyStorage, ICapabilityProvider {
    private var energyStored: Int = 0
    private var speedLevel: Int = 0
    private var fortuneLevel: Int = 0
    private var speedUpgradesNeeded: Int = 1
    private var fortuneUpgradesNeeded: Int = 1
    private var itemsToDrop: MutableList<ItemStack> = ArrayList()
    private var curPos: MutableBlockPos = MutableBlockPos(0, myPos.y - 1, 0)
    private var curChunk: QuarryChunk? = null
    private var chunksTodo: MutableList<QuarryChunk> = ArrayList()

    override fun receiveEnergy(maxReceive: Int, simulate: Boolean): Int {
        val recv = min(QUARRY_MAX_ENERGY - energyStored, maxReceive)
        if (!simulate)
            energyStored += recv
        return recv
    }

    override fun extractEnergy(maxExtract: Int, simulate: Boolean): Int {
        return 0
    }

    override fun getEnergyStored(): Int {
        return energyStored
    }

    override fun getMaxEnergyStored(): Int {
        return QUARRY_MAX_ENERGY
    }

    override fun canExtract(): Boolean {
        return false
    }

    override fun canReceive(): Boolean {
        return true
    }

    override fun <T: Any?> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
        if (cap == CapabilityEnergy.ENERGY) return LazyOptional.of { this as T }
        return super<BlockEntity>.getCapability(cap, side)
    }

    private fun getPickaxe(): ItemStack = Items.NETHERITE_PICKAXE.defaultInstance.apply {
        if (fortuneLevel > 0) enchant(Enchantments.BLOCK_FORTUNE, fortuneLevel)
    }

    private fun breakBlock(level: Level, pos: BlockPos): Boolean {
        val state = level.getBlockState(pos)
        val pickaxe = getPickaxe()
        val player = FakeQuarryPlayer.get(level as ServerLevel)
        player.setItemInHand(InteractionHand.MAIN_HAND, pickaxe)
        val event = BlockEvent.BreakEvent(level, pos, state, player)
        MinecraftForge.EVENT_BUS.post(event)
        if (event.isCanceled) return true
        if (state.isAir) return true
        if (state.getDestroySpeed(level, pos) < 0) return true
        val energyNeeded = ENERGY_BASE + max(0, (state.getDestroySpeed(level, pos) * ENERGY_PER_STRENGTH).toInt())
        if (energyStored < energyNeeded) return false
        energyStored -= energyNeeded
        if (level.getFluidState(pos).isEmpty) // Is not a liquid
            itemsToDrop.addAll(Block.getDrops(state, level, pos, level.getBlockEntity(pos), player, pickaxe))
        level.setBlock(pos, Blocks.AIR.defaultBlockState(), Block.UPDATE_ALL)
        setChanged()
        return true
    }

    private fun newChunk() {
        curChunk = chunksTodo.removeFirstOrNull()
    }

    fun tick(level: Level) {
        // Mine
        if (curChunk == null) newChunk()
        if (itemsToDrop.isEmpty() && curChunk != null) {
            var curChunkPos = curChunk!!.asPos()
            for (i in 0..speedLevel) {
                if (!breakBlock(level, curChunkPos.offset(curPos))) break
                curPos.x++
                if (curPos.x == 16) {
                    curPos.x = 0
                    curPos.z++
                }
                if (curPos.z == 16) {
                    curPos.z = 0
                    curPos.y--
                }
                if (curPos.y == level.minBuildHeight - 1) {
                    curPos.y = myPos.y - 1
                    newChunk()
                    if (curChunk == null) break
                    curChunkPos = curChunk!!.asPos()
                }
            }
        }

        if (itemsToDrop.isEmpty()) return
        // Drop items
        val te = level.getBlockEntity(myPos.offset(0, 1, 0)) ?: return
        val itemHandler = te.getCapability(CapabilityItemHandler.ITEM_HANDLER_CAPABILITY, Direction.DOWN).orElse(null) ?: return
        val newDrops = ArrayList<ItemStack>()
        for (drop in itemsToDrop) {
            val newDrop = ItemHandlerHelper.insertItemStacked(itemHandler, drop, false)
            if (!newDrop.isEmpty) newDrops.add(newDrop)
        }
        itemsToDrop = newDrops
        setChanged()
    }

    fun playerInfo(): String = "Speed: $speedLevel, needed: $speedUpgradesNeeded | Fortune: $fortuneLevel, needed: $fortuneUpgradesNeeded | " + if (curChunk != null) "Chunks in queue: ${chunksTodo.size}" else "Idle"

    fun useUpgrade(player: Player, stack: ItemStack) {
        val item = stack.item as QuarryUpgrade
        if (item.ty == QuarryUpgrade.Type.SPEED) {
            val used = min(stack.count, speedUpgradesNeeded)
            stack.shrink(used)
            speedUpgradesNeeded -= used
            if (speedUpgradesNeeded == 0) {
                speedLevel++
                speedUpgradesNeeded = speedLevel.toFloat().pow(1.5f).toInt()
                player.displayClientMessage(TextComponent("New speed level: $speedLevel, upgrades needed for next level: $speedUpgradesNeeded"), true)
            } else
                player.displayClientMessage(TextComponent("Upgrades needed for next level: $speedUpgradesNeeded"), true)
        } else {
            val used = min(stack.count, fortuneUpgradesNeeded)
            stack.shrink(used)
            fortuneUpgradesNeeded -= used
            if (fortuneUpgradesNeeded == 0) {
                fortuneLevel++
                fortuneUpgradesNeeded = fortuneLevel.toFloat().pow(1.5f).toInt()
                player.displayClientMessage(TextComponent("New fortune level: $fortuneLevel, upgrades needed for next level: $fortuneUpgradesNeeded"), true)
            } else
                player.displayClientMessage(TextComponent("Upgrades needed for next level: $fortuneUpgradesNeeded"), true)
        }
    }

    fun loadFromItem(tag: CompoundTag) {
        speedLevel = tag.getInt("speedLevel")
        fortuneLevel = tag.getInt("fortuneLevel")
        speedUpgradesNeeded = tag.getInt("speedUpgradesNeeded")
        fortuneUpgradesNeeded = tag.getInt("fortuneUpgradesNeeded")
    }

    fun getDropItem(): ItemStack {
        val tag = CompoundTag()
        tag.putInt("speedLevel", speedLevel)
        tag.putInt("fortuneLevel", fortuneLevel)
        tag.putInt("speedUpgradesNeeded", speedUpgradesNeeded)
        tag.putInt("fortuneUpgradesNeeded", fortuneUpgradesNeeded)
        val item = ItemStack(QuarryMod.QUARRY_ITEM.get(), 1)
        item.tag = tag
        return item
    }

    fun addChunk(chunk: QuarryChunk) = chunksTodo.add(chunk)

    private fun saveClientData(tag: CompoundTag) {
        tag.putInt("energy", energyStored)
        tag.putInt("speedLevel", speedLevel)
        tag.putInt("fortuneLevel", fortuneLevel)
        tag.putInt("speedUpgradesNeeded", speedUpgradesNeeded)
        tag.putInt("fortuneUpgradesNeeded", fortuneUpgradesNeeded)
    }

    override fun saveAdditional(tag: CompoundTag) {
        saveClientData(tag)
        val dropsList = ListTag()
        itemsToDrop.stream().forEach {
            val itemTag = CompoundTag()
            it.save(itemTag)
            dropsList.add(itemTag)
        }
        tag.put("drops", dropsList)
        tag.putInt("curPosX", curPos.x)
        tag.putInt("curPosY", curPos.y)
        tag.putInt("curPosZ", curPos.z)
        if (curChunk != null) {
            tag.put("curChunk", curChunk!!.save())
        }
        val chunkList = ListTag()
        chunkList.addAll(chunksTodo.stream().map(QuarryChunk::save).asSequence())
        tag.put("chunks", chunkList)
    }

    override fun load(tag: CompoundTag) {
        energyStored = tag.getInt("energy")
        speedLevel = tag.getInt("speedLevel")
        fortuneLevel = tag.getInt("fortuneLevel")
        speedUpgradesNeeded = tag.getInt("speedUpgradesNeeded")
        fortuneUpgradesNeeded = tag.getInt("fortuneUpgradesNeeded")
        if (!tag.contains("drops")) return // Everything after this is server only
        itemsToDrop = tag.getList("drops", Tag.TAG_COMPOUND.toInt())
                .map { ItemStack.of(it as CompoundTag) }.toMutableList()
        curPos.x = tag.getInt("curPosX")
        curPos.y = tag.getInt("curPosY")
        curPos.z = tag.getInt("curPosZ")
        curChunk = if (tag.contains("curChunk")) QuarryChunk(tag.get("curChunk") as CompoundTag) else null
        chunksTodo = tag.getList("chunks", CompoundTag.TAG_COMPOUND.toInt())
                .map { QuarryChunk(it as CompoundTag) }.toMutableList()
    }

    override fun getUpdateTag(): CompoundTag {
        val tag = CompoundTag()
        saveClientData(tag)
        return tag
    }

    override fun getUpdatePacket(): Packet<ClientGamePacketListener>? = ClientboundBlockEntityDataPacket.create(this)

    fun probeInfo(info: IProbeInfo) {
        val speedMax = (2.0).pow(speedLevel).toInt()
        info.text("Speed: level $speedLevel, ${speedMax - speedUpgradesNeeded}/$speedMax")
        val fortuneMax = (2.0).pow(fortuneLevel).toInt()
        info.text("Fortune: level $fortuneLevel, ${fortuneMax - fortuneUpgradesNeeded}/$fortuneMax")
        if (itemsToDrop.isNotEmpty())
            info.text("Waiting to extract items...")
        else if (curChunk != null) {
            val pos = curChunk!!.asPos().offset(curPos)
            info.text("Currently mining: ${pos.x} ${pos.y} ${pos.z}")
            info.text("Chunks in queue: ${chunksTodo.size}")
        } else info.text("Idle")
    }

    companion object {
        const val QUARRY_MAX_ENERGY: Int = 100_000_000
        const val ENERGY_BASE = 2000
        const val ENERGY_PER_STRENGTH = 1000
    }
}